@@ -6,7 +6,7 @@
 #' @param det distribution element tree object resulting from \code{\link{det.construct}}.
 #' @param xc vector with conditioning values of probability-space components listed in \code{dc}. If empty (default), unconditional samples are generated.
 #' @param dc integer vector with indices of conditioning components corresponding to \code{xc}. If empty (default), unconditional samples are generated.
-#' @param cores for large \code{n}, \code{cores > 1} allows for parallel bootstrapping. The default is \code{cores = 0}, which allocates half of the available cores (see \code{\link[parallel]{detectCores}}). \code{cores = 1} corresponds to serial bootstrapping.
+#' @param cores for large \code{n}, \code{cores > 1} allows for parallel bootstrapping using the indicated number of cores. The default is \code{cores = Inf}, which allocates half of the available cores (see \code{\link[parallel]{detectCores}}). \code{cores = 1} corresponds to serial bootstrapping.
 #'
 #' @return A matrix containing \code{n} random vectors (columns) with \code{d} components or dimensions (rows) is returned. \code{d} is equal to the dimensionality of the underlying \code{det} object.
 #' @export
@@ -22,45 +22,42 @@
 #' plot(x, type = "p", pch = ".", asp = 1)
 #' lines(c(0,1), x2c*c(1,1), col = "red")
 #' # DET construction and bootstrapping
-#' det <- det.construct(t(x), mode = 1, lb = 0, ub = 0, cores = 1) # const. de's, no pre-white
-#' y <- det.rnd(1e3, det, xc = x2c, dc = 2, cores = 1) # conditional bootstrap'g
+#' det <- det.construct(t(x), mode = 1, lb = 0, ub = 0) # const. de's, no pre-white
+#' y <- det.rnd(1e3, det, xc = x2c, dc = 2, cores = 2) # conditional bootstrap'g
 #' # compare generated data (black) with exact cond. distribution (red)
-#' screen(2); det1(y[1,], mode = 1, cores = 2)
+#' screen(2); det1(y[1,], mode = 1)
 #' lines(c(0,x2c,x2c,1,1),c(0,0,1/(1-x2c),1/(1-x2c),0), col = "red")
 #'
-#' grDevices::dev.off() # erase plot
-#'
 #' ## example 2d unconditional
 #' require(stats); require(graphics)
-#' x <- matrix(runif(1e4), ncol = 2); x <- x[x[,2]<x[,1],] # uniform wedge
-#' det <- det.construct(t(x), mode = 1, lb = 0, ub = 0, cores = 1) # no pre-white
-#' y <- t(det.rnd(nrow(x), det, cores = 1)) # smooth bootstrapping
+#' x <- matrix(runif(2e4), ncol = 2); x <- x[x[,2]<x[,1],] # uniform wedge
+#' det <- det.construct(t(x), mode = 1, lb = 0, ub = 0) # no pre-white
+#' y <- t(det.rnd(nrow(x), det, cores = 2)) # smooth bootstrapping
 #' split.screen(c(2, 1))
 #' screen(1); plot(x, type = "p", pch = ".", asp = 1, main = "original")
 #' screen(2); plot(y, type = "p", pch = ".", asp = 1, main = "bootstrapped")
 #'
-#' grDevices::dev.off() # erase plot
-#'
 #' ## example 3d
 #' require(stats); require(graphics)
 #' # mean and covariance of Gaussian, data generation
 #' mu <- c(1,3,2); C <- matrix(c(25,7.5,1.75,7.5,7,1.35,1.75,1.35,0.43), nrow = 3)
 #' A <- eigen(C); B <- diag(A$values); A <- A$vectors
-#' x <- matrix(rnorm(24e3), nrow = 3)
+#' x <- matrix(rnorm(3e4), nrow = 3)
 #' x <- A %*% (sqrt(B) %*% x) + mu %*% t(rep(1,ncol(x)))
 #' lbl <- "x1 | x2 = 7 & x3 = 2.5"
 #' pairs(t(x), labels = c("x1","x2","x3"), pch = ".", main = lbl)
 #' # bootstrapping conditional on x2 and x3
-#' det <- det.construct(x, lb = 0, ub = 0, cores = 2)
+#' det <- det.construct(x, lb = 0, ub = 0)
 #' xc <- c(2.5,7); d <- c(3,2) # conditional on x2 = 7 & x3 = 2.5
-#' y <- det.rnd(1e3, det, xc, d, cores = 2)
-#' det1(y[1,], mode = 1, main = lbl, cores = 2)
+#' y <- det.rnd(1e4, det, xc, d, cores = 1)
+#' det1(y[1,], mode = 1, main = lbl)
 #' # compare with exact conditional density
 #' Cm1 <- solve(C); var1 <- det(C)/det(C[2:3,2:3]) # conditional variance
 #' mu1 <- mu[1] + var1*((mu[2]-xc[d==2])*Cm1[1,2]+(mu[3]-xc[d==3])*Cm1[1,3]) # cond. mean
 #' x1 <- mu1 + seq(-50,50)/50 * 5*sqrt(var1) # x1-axis grid points
 #' lines(x1, dnorm(x1,mu1,sqrt(var1)), col = "red")
-det.rnd <- function(n, det, xc = vector("numeric", length = 0), dc = vector("numeric", length = 0), cores = 0) {
+det.rnd <- function(n, det, xc = vector("numeric", length = 0),
+                            dc = vector("numeric", length = 0), cores = Inf) {
    ds <- length(det$lb) # number of sample-space dimensions
    # collect det leafs cut by condit'g planes p
    if (length(xc) == 0) { # unconditional case
@@ -93,17 +90,16 @@ det.rnd <- function(n, det, xc = vector("numeric", length = 0), dc = vector("num
    sl <- matrix(rep(0,ds*m), nrow = ds) # slope parameter
    for (k in 1:m) {if (!is.na(theta[k])) {sl[,k] <- theta[[k]]}}
    p <- p * apply(size, 2, prod) *
-      apply((((xc%*%t(rep(1,m))-lb[dc,])/size[dc,] - 1/2)*sl[dc,] + 1) / size[dc,], 2, prod)
-   # setup cluster for parallel processing
-   if (cores <= 0) {cores <- max(1,ceiling(parallel::detectCores()/2))}
-   clster <- parallel::makeCluster(cores)
-   # generate random samples
+      apply((((xc%*%t(rep(1,m))-lb[dc,])/size[dc,] - 1/2)*sl[dc,] + 1)
+            / size[dc,], 2, prod)
    p <- cumsum(p)
-   if (p[m] == 0) {return(NULL)} # zero probability
+   if (p[m] == 0) {return(NULL)} # zero probability -> no samples
    p <- c(0,p/p[m]) # normalize probabilities
+   # list with uniform random numbers (element selection, random components)
    rl <- stats::runif(n*(ds-length(dc)+1))
    rl <- matrix(rl, ncol = n); rl[1,rl[1,] == 0] <- 1
-   x <- parallel::parCapply(clster, rl, function(r) {
+   # generate one random sample from det
+   rdet <- function(r) {
       x <- rep(0,ds)
       ind <- seq(m)[(p[1:m] < r[1]) & (r[1] <= p[2:(m+1)])]
       # randomly draw sample inside selected leaf with inverse marginal cdfs
@@ -126,8 +122,16 @@ det.rnd <- function(n, det, xc = vector("numeric", length = 0), dc = vector("num
          }
       }
       return(x)
-   })
-   parallel::stopCluster(clster)
+   }
+   # generate random samples
+   if (cores > 1) { # parallel processing
+      if (is.infinite(cores)) {cores <- max(1,ceiling(parallel::detectCores()/2))}
+      clster <- parallel::makeCluster(cores)
+      x <- parallel::parCapply(clster, rl, rdet)
+      parallel::stopCluster(clster)
+   } else { # serial
+      x <- apply(rl, MARGIN = 2, rdet)
+   }
    x <- matrix(x, nrow = ds)
    # in unconditional case, transform samples back
    if (length(xc) == 0) {x <- t(det$A)%*%x + det$mu%*%t(rep(1,n))}
@@ -150,7 +154,7 @@ det.rnd <- function(n, det, xc = vector("numeric", length = 0), dc = vector("num
 #' require(stats); require(graphics)
 #' n <- 8e4; x <- rnorm(n)
 #' x <- matrix(c(x, x+rnorm(n,0,0.2)), ncol = 2)
-#' det <- det.construct(t(x), lb = 0, ub = 0, cores = 2) # no pre-whitening
+#' det <- det.construct(t(x), lb = 0, ub = 0) # no pre-whitening
 #' plot(x, type = "p", pch = ".", asp = 1)
 #' # leaf elements that are cut by x1 = 2
 #' leafs <- det.cut(det, xc = 2, dc = 1) # condition x1 = 2
